#include "ASynchronizationObjectLock.hpp"
#include "ASynchronizationObject.hpp"
#include "ATrace.hpp"

// Copyright (c) 1997-2000 Askesis B.V. See licence.txt for details.
// For information, bug reports and additions send e-mail to ALib@askesis.nl

// Author : Joost Kraaijeveld
// Created: 18 April 2000 08:21:00

// public
// Constructors
ASynchronizationObjectLock::ASynchronizationObjectLock(ASynchronizationObject& aASynchronizationObject, long aTimeOut)
{
   synchronizationObject = &aASynchronizationObject;
   lock(aTimeOut);
}

// Destructor
ASynchronizationObjectLock::~ASynchronizationObjectLock()
{
   if(synchronizationObject && synchronizationObject->referenceCounter)
   {
      unlock();
   }
}

// Operators

// Accessors

// Mutators

// Operations

// protected
// Operations
void ASynchronizationObjectLock::lock(long aTimeOut)
{
   if (synchronizationObject->getHandle().isValid())
   {
      if (APlatform::isWin32S())
      {
         ATRACE_DEVELOP("Win32s: ASynchronizationObjectLock::lock(long aTimeOut) does not wait on a semaphore");
      }else
      {
         // The WaitForSingleObject function returns when the specified object is in
         // the signaled state or when the time-out interval elapses.

         // Return Value

         // If the function succeeds, the return value indicates the event that caused
         // the function to return.
         // If the function fails, the return value is WAIT_FAILED. To get extended
         // error information, call GetLastError.

         // The return value on success is one of the following values:

         // WAIT_ABANDONED:   The specified object is a mutex object that was not released
         //                   by the thread that owned the mutex object before the owning
         //                   thread terminated. Ownership of the mutex object is granted
         //                   to the calling thread, and the mutex is set to nonsignaled.
         // WAIT_OBJECT_0:    The state of the specified object is signaled.
         // WAIT_TIMEOUT:     The time-out interval elapsed, and the object's state is nonsignaled.

         //Remarks

         // The WaitForSingleObject function checks the current state of the specified object.
         // If the object's state is nonsignaled, the calling thread enters an efficient wait
         // state. The thread consumes very little processor time while waiting for one of the
         // conditions of the wait to be satisfied.
         // The system adjusts the state of some types of objects after the object's signaled
         // state satisfies a wait condition. For example, the count of a semaphore object
         // is decreased by one.
         unsigned long result = WaitForSingleObject( synchronizationObject->getHandle(), aTimeOut);
         if (result == WAIT_FAILED)
         {
            ALibBaseException exception("WaitForSingleObject WAIT_FAILED");
            ATRACE_DEVELOP(exception.text());
            ATHROW(exception);
         }else
         {  // Success
            switch(result)
            {
               case WAIT_OBJECT_0:
               {
                  // The state of the specified object is signaled.
                  // This should be the normal case
                  ATRACE_DEVELOP("WaitForSingleObject WAIT_OBJECT_0");
                  break;
               }
               case WAIT_ABANDONED:
               {
                  // The specified object is a mutex object that was not released by the thread that owned the
                  // mutex object before the owning thread terminated. Ownership of the mutex object is granted
                  // to the calling thread, and the mutex is set to nonsignaled.
                  ATRACE_DEVELOP("WaitForSingleObject ABANDONED");
                  break;
               }
               case WAIT_TIMEOUT:
               {
                  // The time-out interval elapsed, and the object's state is nonsignaled.
                  ALibBaseException exception("WaitForSingleObject WAIT_TIMEOUT");
                  ATRACE_DEVELOP(exception.text());
                  ATHROW(exception);
                  break;
               }
               default:
               {
                  ATRACE_DEVELOP("WaitForSingleObject UNKNOWN RESULT");
                  break;
               }
            }
         }
      }
   }
   // According to IBM source code this depends on the processor ????
   // Critical section??
   synchronizationObject->referenceCounter++;
}

void ASynchronizationObjectLock::unlock()
{
   // IBM Source code:
   // DEFECT 24079 : must decrement count while
   //               we still have the mutex.
   synchronizationObject->referenceCounter--;

   if (synchronizationObject->getHandle().isValid())
   {
      if (APlatform::isWin32S())
      {
         ATRACE_DEVELOP("Win32s: ASynchronizationObjectLock::unlock() does not release semaphore");
      }else
      {
         if ( !ReleaseMutex( synchronizationObject->getHandle()))
         {
            // IBM Source code:
            // DEFECT 24079 : should only drop ref count if
            //                semaphore release was successful
            //                Must re-increment it if release failed.
            synchronizationObject->referenceCounter++;
            ALibBaseException exception("ReleaseMutex");
            ATHROW(exception);
         }
      }
   }
}

// private
// Operations
ASynchronizationObjectLock::ASynchronizationObjectLock( const ASynchronizationObjectLock& aASynchronizationObjectLock)
{
}

ASynchronizationObjectLock& ASynchronizationObjectLock::operator=(const ASynchronizationObjectLock& aASynchronizationObjectLock)
{
   return *this;
}
