#include "ATrace.hpp"
#include "ACriticalSectionObject.hpp"  // Do not use any aTrace statements in the ACriticalSectionObject class
#include "AHandle.hpp"                 // Idem
#include "ATraceThreadIndention.hpp"   // Idem

/* static */ bool ATrace::traceOn = false;            // Default no tracing, notice the initialize function thought.
/* static */ bool ATrace::traceLineNumber = true;     // If tracing default to linenumber
/* static */ bool ATrace::traceThreadId = true;
/* static */ bool ATrace::traceInitialized = false;

class ATraceMutexLock;

class ATraceMutex
{
   public:
      ATraceMutex();
      ~ATraceMutex();
      void                                      lock( long aTimeOut = -1);
      void                                      unlock();

      AHandle                                   semaphoreHandle;
      unsigned long                             referenceCounter;
      ATraceMutexLock*                          traceMutexLock;
};

class ATraceMutexLock
{
   public:
      ATraceMutexLock(ATraceMutex& aATraceMutex, long aTimeOut = -1);
      ~ATraceMutexLock();
      void                                      lock(long aTimeOut = -1);
      void                                      unlock();

      ATraceMutex*                              traceMutex;
};

ATraceMutex::ATraceMutex()
   :  semaphoreHandle(0),
      referenceCounter(0),
      traceMutexLock(0)
{
   printf("+ATraceMutex::ATraceMutex()\n");;
   semaphoreHandle = CreateMutex( 0, false, 0 );
   if(!semaphoreHandle.isValid())
   {
      printf("CreateMutex\n");
      ALibBaseException exception("CreateMutex");
      ATHROW(exception);
   }
   printf("-ATraceMutex::ATraceMutex()\n");;
}

ATraceMutex::~ATraceMutex()
{
   printf("+ATraceMutex::~ATraceMutex()\n");
   if (semaphoreHandle.isValid())
   {
      if ( !CloseHandle( semaphoreHandle ) )
      {
         printf("CloseHandle\n");
         ALibBaseException exception("CloseHandle");
         ATHROW(exception);
      }
   }
   semaphoreHandle = 0;

   if(traceMutexLock)
   {
      delete traceMutexLock;
      traceMutexLock = 0;
   }
   referenceCounter = 0;
   printf("-ATraceMutex::~ATraceMutex()\n");
}


void ATraceMutex::lock( long aTimeOut)
{
   if(traceMutexLock==0)
   {
      traceMutexLock = new ATraceMutexLock(*this,aTimeOut);
   }else
   {
      traceMutexLock->lock(aTimeOut);
   }
}

void ATraceMutex::unlock()
{
   if(traceMutexLock!=0)
   {
      traceMutexLock->unlock();
   }
}

ATraceMutexLock::ATraceMutexLock(ATraceMutex& aATraceMutex, long aTimeOut)
{
   traceMutex = &aATraceMutex;
   lock(aTimeOut);
}

ATraceMutexLock::~ATraceMutexLock()
{
   if(traceMutex && traceMutex->referenceCounter)
   {
      unlock();
   }
}

void ATraceMutexLock::lock(long aTimeOut)
{
   if (traceMutex->semaphoreHandle.isValid())
   {
      // The WaitForSingleObject function returns when the specified object is in
      // the signaled state or when the time-out interval elapses.

      // Return Value

      // If the function succeeds, the return value indicates the event that caused
      // the function to return.
      // If the function fails, the return value is WAIT_FAILED. To get extended
      // error information, call GetLastError.

      // The return value on success is one of the following values:

      // WAIT_ABANDONED:   The specified object is a mutex object that was not released
      //                   by the thread that owned the mutex object before the owning
      //                   thread terminated. Ownership of the mutex object is granted
      //                   to the calling thread, and the mutex is set to nonsignaled.
      // WAIT_OBJECT_0:    The state of the specified object is signaled.
      // WAIT_TIMEOUT:     The time-out interval elapsed, and the object's state is nonsignaled.

      //Remarks

      // The WaitForSingleObject function checks the current state of the specified object.
      // If the object's state is nonsignaled, the calling thread enters an efficient wait
      // state. The thread consumes very little processor time while waiting for one of the
      // conditions of the wait to be satisfied.
      // The system adjusts the state of some types of objects after the object's signaled
      // state satisfies a wait condition. For example, the count of a semaphore object
      // is decreased by one.
      unsigned long result = WaitForSingleObject( traceMutex->semaphoreHandle, aTimeOut );
      if (result == WAIT_FAILED)
      {
         printf("ATraceMutexLock: WaitForSingleObject WAIT_FAILED");
         ALibBaseException exception("WaitForSingleObject WAIT_FAILED");
         ATHROW(exception);
      }else
      {  // Success
         switch(result)
         {
            case WAIT_OBJECT_0:
            {
               // The state of the specified object is signaled.
               // This should be the normal case
//               printf("ATraceMutexLock: WaitForSingleObject WAIT_OBJECT_0\n");
               break;
            }
            case WAIT_ABANDONED:
            {
               // The specified object is a mutex object that was not released by the thread that owned the
               // mutex object before the owning thread terminated. Ownership of the mutex object is granted
               // to the calling thread, and the mutex is set to nonsignaled.
//               printf("ATraceMutexLock: WaitForSingleObject ABANDONED\n");
               break;
            }
            case WAIT_TIMEOUT:
            {
               // The time-out interval elapsed, and the object's state is nonsignaled.
               printf("ATraceMutexLock:WaitForSingleObject WAIT_TIMEOUT");
               ALibBaseException exception("WaitForSingleObject WAIT_TIMEOUT");
               ATHROW(exception);
               break;
            }
            default:
            {
               printf("ATraceMutexLock: WaitForSingleObject UNKNOWN RESULT\n");
               break;
            }
         }
      }
      // According to IBM source code this depends on the processor ????
      // Critical section??
      {
         ACriticalSectionObject criticalSection;
         traceMutex->referenceCounter++;
      }
   }else
   {
      // Semaphore handle is not valid.
      printf("void ATraceMutexLock::lock(): trying to lock invalid ATraceMutex\n");
      ALibBaseException exception("void ATraceMutexLock::lock(): trying to lock invalid ATraceMutex");
      ATHROW(exception);
   }
}

void ATraceMutexLock::unlock()
{
   if (traceMutex->semaphoreHandle.isValid())
   {
      // IBM Source code:
      // DEFECT 24079 : must decrement count while
      //               we still have the mutex.
      {
         ACriticalSectionObject criticalSection;
         traceMutex->referenceCounter--;
      }

      if ( !ReleaseMutex( traceMutex->semaphoreHandle))
      {
         // IBM Source code:
         // DEFECT 24079 : should only drop ref count if
         //                semaphore release was successful
         //                Must re-increment it if release failed.
         {
            ACriticalSectionObject criticalSection;
            traceMutex->referenceCounter++;
         }

         printf("ATraceMutexLock: ReleaseMutex()\n");
         ALibBaseException exception("ReleaseMutex");
         ATHROW(exception);
      }
   }else
   {
      // Semaphore handle is not valid.
      printf("void ATraceMutexLock::unlock(): trying to unlock invalid ATraceMutex\n");
      ALibBaseException exception("void ATraceMutexLock::unlock(): trying to unlock invalid ATraceMutex");
      ATHROW(exception);
   }
}

//ATraceMutex traceMutex;

// public
// Constructors
ATrace::ATrace(const char* aText,
               long aLineNumber,
               bool withTimeStamp)
   :  traceTimeStamp(withTimeStamp)
{
   if(ATrace::isTraceOn())
   {
      if (aText!=0)
      {
         string timeStampString("");
         traceText = aText;
         string lineNumberString("");

         if(traceTimeStamp==true)
         {
            timeStampString = ATrace::timeStamp();
            timeStampString += ": ";
         }
         if(ATrace::isLineNumberEnabled() && aLineNumber>0)
         {
            lineNumberString = " (line ";
            char numberBuffer[10+1]; // max long ==__LONG_MAX__ == 2147483647 == 10 chars
            sprintf(&numberBuffer[0],"%ld",aLineNumber);
            lineNumberString+=string(&numberBuffer[0]);
            lineNumberString+=")";
         }
         string formattedString = timeStampString + traceText + lineNumberString;
         ATrace::traceFormattedString( formattedString, "+");
      }else
      {
         ATrace::trace("", "+");
      }
   }
}

// Destructor
ATrace::~ATrace()
{
   if(ATrace::isTraceOn())
   {
      string timeStampString;

      // The timestamp is printed before the trace text with ":" appended
      if(traceTimeStamp==true)
      {
         timeStampString = ATrace::timeStamp();
         timeStampString += ": ";
      }

      // Print the actual trace text
      string formattedString = timeStampString + traceText;
      ATrace::traceFormattedString( formattedString, "-");
   }
}

// Operators

// Accessors

// Mutators

// Operations
/* static */ void ATrace::enableTracing()
{
   ACriticalSectionObject criticalSection;
   traceOn = true;
}

/* static */ void ATrace::disableTracing()
{
   ACriticalSectionObject criticalSection;
   traceOn = false;
}

/* static */ bool ATrace::isTraceEnabled()
{
   return traceOn;
}

/* static */ void ATrace::enableLinenumber()
{
   ACriticalSectionObject criticalSection;
   traceLineNumber = true;
}

/* static */ void ATrace::disableLinenumber()
{
   ACriticalSectionObject criticalSection;
   traceLineNumber = false;
}

/* static */ bool ATrace::isLineNumberEnabled()
{
   return traceLineNumber;
}

/* static */ void ATrace::enableThreadId()
{
   ACriticalSectionObject criticalSection;
   traceThreadId = true;
}

/* static */ void ATrace::disableThreadId()
{
   ACriticalSectionObject criticalSection;
   traceThreadId = false;
}

/* static */ bool ATrace::isThreadIdEnabled()
{
   return traceThreadId;
}

static ATraceMutex traceMutex;

/* static */ void ATrace::trace( const string& aText,
                                 bool withTimeStamp)
{
   // Lock this function. Locking this function also protects threadIndentionLevels
   // as this is the only function it is used.
   ATraceMutexLock traceMutexLock(traceMutex);

   if(ATrace::isTraceOn())
   {
      string timeStampString;

      // The timestamp is printed before the trace text with ":" appended
      if(withTimeStamp==true)
      {
         timeStampString = ATrace::timeStamp();
         timeStampString += ": ";
      }

      // Print the actual trace text
      string formattedString = timeStampString + aText;
      ATrace::traceFormattedString( formattedString, ">");
   }
}

/* static */ void ATrace::trace( const char* aText,
                                 bool withTimeStamp)
{
   string tempString(aText);
   ATrace::trace(tempString,withTimeStamp);
}

/* static */ void ATrace::traceFormattedString( const string& aText,
                                                char* aMarker)
{
   // Lock this function. Locking this function also protects threadIndentionLevels
   // as this is the only function it is used.
   ATraceMutexLock traceMutexLock(traceMutex);

   //threadIndentionLevels
   // If the application is multithreaded the indentionlevel is maintained per thread
   static AThreadIndentionLevels threadIndentionLevels;

   AThreadId currentThreadId = GetCurrentThreadId();
   string margin;

   switch(*aMarker)
   {
      case '+':
      {
         // First indent, than print
         margin.insert((unsigned int)0,(++threadIndentionLevels[currentThreadId]).indentionLevel,' ');
         break;
      }
      case '-':
      {
         // First print than "out"dent
         margin.insert((unsigned int)0,(threadIndentionLevels[currentThreadId]--).indentionLevel,' ');
         break;
      }
      default:
      {
         margin.insert((unsigned int)0,threadIndentionLevels[currentThreadId].indentionLevel + 3,' ');
         break;
      }
   }

   if(ATrace::isThreadIdEnabled())
   {
      printf("%s%lu: %c%s\n",margin.c_str(),currentThreadId.asUnsigned(),*aMarker,aText.c_str());
   }else
   {
      printf("%s%c%s\n",margin.c_str(),*aMarker,aText.c_str());
   }
}

/* static */ const char* ATrace::timeStamp()
{
   // TODO: Returning a static buffer in a multithread environment is a bad idea. Got to change that someday.
   static char timeStamp[27];

   struct timeb timeBuffer;
   ftime(&timeBuffer);
   struct tm* timeStruct = localtime(&timeBuffer.time);
   sprintf(timeStamp,"%04d-%02d-%02d-%02d-%02d-%02d-%06d",timeStruct->tm_year+1900,timeStruct->tm_mon+1,timeStruct->tm_mday,timeStruct->tm_hour,timeStruct->tm_min,timeStruct->tm_sec,timeBuffer.millitm);
   return &timeStamp[0];
}

/* static */ bool ATrace::isTraceOn()
{
   ACriticalSectionObject criticalSection;

   if( ATrace::traceInitialized == false)
   {
      ATrace::traceOn            = true;  // TODO: check the environment/configuration file for settings
      ATrace::traceLineNumber    = true;  // TODO: check the environment/configuration file for settings
      ATrace::traceThreadId      = true;  // TODO: check the environment/configuration file for settings
      ATrace::traceInitialized   = true;
   }
   return ATrace::traceOn;
}

// protected
// Operations

// private
// Operations

// End of class definition of ATrace
