#ifndef _ATrace_hpp_
#define _ATrace_hpp_

// Copyright (c) 1997-2000 Askesis B.V. See licence.txt for details.
// For information, bug reports and additions send e-mail to ALib@askesis.nl

// Author : Joost Kraaijeveld
// Created: 26 Nov 1997 08:29:08
// Purpose: Trace class for debugging purposes.
//
//          If an object of this class is instantiated a trace is printed in the ctor and dtor.
//          An object is meaned for function tracing: instantiate the object as first
//          statement in the function (which will trace a text) and let it go out of scope
//          at the end of the function (which will also a text).
//
//          If just a single line needs to be printed, use one of the static trace functions.
//
//          Normally, use the macros. They know.
//
//          Implementation notes: Do not use any class that might trace to implement ATrace.
//          This leadt to circular stuff. At this moment ACriticalSectionObject is used.
//          As soon as we have a good debugged AMutexObject that will be used also.

// Configuration file
#include "ALibConfig.h"

// Standard C/C++ library headers
#include <iostream>
#include <sys/timeb.h>  // _ftime(), time_t and struct timeb
#include <time.h>       // localtime()
#include <stdio.h>      // sprintf()

// Operating system toolkit headers

// STL headers
#include <string>

// Other class headers

// Forward declarations

class ALIBIMPEXP ATrace
{
   public:
      // Constructors
      ATrace(  const char* aText    = 0,
               long aLineNumber     = 0,
               bool withTimeStamp   = false);

      // Destructor
      virtual ~ATrace();

      // Operators

      // Accessors

      // Mutators

      // Operations
      // ATrace functions
      static void                               enableTracing();
      static void                               disableTracing();
      static bool                               isTraceEnabled();

      static void                               enableLinenumber();
      static void                               disableLinenumber();
      static bool                               isLineNumberEnabled();

      static void                               enableThreadId();
      static void                               disableThreadId();
      static bool                               isThreadIdEnabled();

      static void                               trace(const string& aText,
                                                      bool withTimeStamp = false);
      static void                               trace(const char* aText,
                                                      bool withTimeStamp = false);
      // The two functions above resolve to this function after formatting the text
      static void                               traceFormattedString(const string& aText,
                                                                     char* aMarker);      // This text will be prepended
                                                                                          // After the text a newline will be inserted
                                                                                          // as approriate
      // Attributes

   protected:
      // Attributes

      // Operations
      static const char*                        timeStamp();
      static bool                               isTraceOn();

   private:
      // Friends

      // Operations

      // Attributes
      static bool                               traceOn;
      static bool                               traceLineNumber;
      static bool                               traceThreadId;
      static bool                               traceInitialized;
      string                                    traceText;           // Needed to trace text in both ctor and dtor
      bool                                      traceTimeStamp;      // Needed to trace timestamp in both ctor and dtor

}; // End of class declaration of ATrace


#ifdef ALIB_TRACE_ALL

   #define AMODTRACE_ALL(modname)               ATrace trc(modname, __LINE__)

#ifdef __IBMCPP__
   #define AFUNCTRACE_ALL()                     ATrace trc(__FUNCTION__, __LINE__)
#endif
#if defined(__CYGWIN32__) || defined(__MINGW32__)
   #define AFUNCTRACE_ALL()                     ATrace trc(__PRETTY_FUNCTION__, __LINE__)
#endif

   #define ATRACE_ALL(text)                     ATrace::trace(text)

   #define ATSMODTRACE_ALL(modname)             ATrace trc(modname, __LINE__, true)

#ifdef __IBMCPP__
   #define ATSFUNCTRACE_ALL()                   ATrace trc(__FUNCTION__, __LINE__)
#endif
#if defined(__CYGWIN32__) || defined(__MINGW32__)
   #define ATSFUNCTRACE_ALL()                   ATrace trc(__PRETTY_FUNCTION__, __LINE__)
#endif

   #define ATSTRACE_ALL(text)                   ATrace::trace(text,true)

   #ifndef ALIB_TRACE_DEVELOP
      #define ALIB_TRACE_DEVELOP
   #endif
#else
   #define AMODTRACE_ALL(modname)
   #define AFUNCTRACE_ALL()
   #define ATRACE_ALL(text)

   #define ATSMODTRACE_ALL(modname)
   #define ATSFUNCTRACE_ALL()
   #define ATSTRACE_ALL(text)
#endif

#ifdef ALIB_TRACE_DEVELOP
   #define AMODTRACE_DEVELOP(modname)           ATrace trc(modname,  __LINE__)

#ifdef __IBMCPP__
   #define AFUNCTRACE_DEVELOP()                 ATrace trc(__FUNCTION__, __LINE__)
#endif
#if defined(__CYGWIN32__) || defined(__MINGW32__)
   #define AFUNCTRACE_DEVELOP()                 ATrace trc(__PRETTY_FUNCTION__, __LINE__)
#endif

   #define ATRACE_DEVELOP(text)                 ATrace::trace(text)

   #define ATSMODTRACE_DEVELOP(modname)         ATrace trc(modname,__LINE__,true)

#ifdef __IBMCPP__
   #define ATSFUNCTRACE_DEVELOP()               ATrace trc(__FUNCTION__, __LINE__)
#endif
#if defined(__CYGWIN32__) || defined(__MINGW32__)
   #define ATSFUNCTRACE_DEVELOP()               ATrace trc(__PRETTY_FUNCTION__, __LINE__)
#endif

   #define ATSTRACE_DEVELOP(text)               ATrace::trace(text,true)

   #ifndef ALIB_TRACE_RUNTIME
      #define ALIB_TRACE_RUNTIME
   #endif
#else
   #define AMODTRACE_DEVELOP(modname)
   #define AFUNCTRACE_DEVELOP()
   #define ATRACE_DEVELOP(text)

   #define ATSMODTRACE_DEVELOP(modname)
   #define ATSFUNCTRACE_DEVELOP()
   #define ATSTRACE_DEVELOP(text)
#endif

#ifdef ALIB_TRACE_RUNTIME
   #define AMODTRACE_RUNTIME(modname)           ATrace trc(modname, __LINE__ )

#ifdef __IBMCPP__
   #define AFUNCTRACE_RUNTIME()                 ATrace trc(__FUNCTION__, __LINE__)
#endif
#if defined(__CYGWIN32__) || defined(__MINGW32__)
   #define AFUNCTRACE_RUNTIME()                 ATrace trc(__PRETTY_FUNCTION__, __LINE__)
#endif

   #define ATRACE_RUNTIME(text)                 ATrace::trace(text)

   #define ATSMODTRACE_RUNTIME(modname)         ATrace trc(modname,__LINE__,true)

#ifdef __IBMCPP__
   #define ATSFUNCTRACE_RUNTIME()               ATrace trc(__FUNCTION__, __LINE__)
#endif
#if defined(__CYGWIN32__) || defined(__MINGW32__)
   #define ATSFUNCTRACE_RUNTIME()               ATrace trc(__PRETTY_FUNCTION__, __LINE__)
#endif

   #define ATSTRACE_RUNTIME(text)               ATrace::trace(text,true)

#else
   #define AMODTRACE_RUNTIME(modname)
   #define AFUNCTRACE_RUNTIME()
   #define ATRACE_RUNTIME(text)

   #define ATSMODTRACE_RUNTIME(modname)
   #define ATSFUNCTRACE_RUNTIME()
   #define ATSTRACE_RUNTIME(text)
#endif

#endif
